# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" Finetuning the library models for multiple choice (Bert, Roberta, XLNet)."""


import os
import sys
import json
import logging


from typing import Dict, Optional, List, NamedTuple
from collections import defaultdict
from dataclasses import dataclass, field

import numpy as np

from transformers import (
    AutoConfig,
    AutoModelForMultipleChoice,
    AutoTokenizer,
    EvalPrediction,
    HfArgumentParser,
    Trainer,
    TrainingArguments,
    set_seed,
)
from transformers import is_tf_available
from transformers.trainer_utils import PredictionOutput
from mc_transformers.utils_mc import MultipleChoiceDataset, Split, processors
from mc_transformers.data_classes import InputExample


if is_tf_available():
    # Force no unnecessary allocation
    import tensorflow as tf
    gpus = tf.config.experimental.list_physical_devices('GPU')
    for gpu in gpus:
        tf.config.experimental.set_memory_growth(gpu, True)


logger = logging.getLogger(__name__)
os.environ.update(**{"WANDB_DISABLED": "true"})


def simple_accuracy(preds, labels):
    return (preds == labels).mean()


def softmax(preds, axis=None):
    # Taken from: https://nolanbconaway.github.io/blog/2017/softmax-numpy.html
    if axis is None:
        raise ValueError("Softmax function needs an axis to work!")
    # make preds at least 2d
    y = np.atleast_2d(preds)
    # subtract the max for numerical stability
    y = y - np.expand_dims(np.max(y, axis=axis), axis)
    y = np.exp(y)
    # take the sum along the specified axis
    ax_sum = np.expand_dims(np.sum(y, axis=axis), axis)
    p = y / ax_sum
    # flatten if preds was 1D
    if len(preds.shape) == 1:
        p = p.flatten()

    return p


@dataclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """

    model_name_or_path: str = field(
        metadata={"help": "Path to pretrained model or model identifier from huggingface.co/models"}
    )
    config_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained config name or path if not the same as model_name"}
    )
    tokenizer_name: Optional[str] = field(
        default=None, metadata={"help": "Pretrained tokenizer name or path if not the same as model_name"}
    )
    cache_dir: Optional[str] = field(
        default=None, metadata={"help": "Where do you want to store the pretrained models downloaded from s3"}
    )


@dataclass
class DataTrainingArguments:
    """
    Arguments pertaining to what data we are going to input our model for training and eval.
    """

    task_name: str = field(metadata={"help": "The name of the task to train on: " + ", ".join(processors.keys())})
    data_dir: str = field(metadata={"help": "Should contain the data files for the task."})
    max_seq_length: int = field(
        default=128,
        metadata={
            "help": "The maximum total input sequence length after tokenization. Sequences longer "
            "than this will be truncated, sequences shorter will be padded."
        },
    )
    overwrite_cache: bool = field(
        default=False, metadata={"help": "Overwrite the cached training and evaluation sets"}
    )


@dataclass
class DirArguments:
    """
    Arguments pertaining to output directories for metrics, results and predictions
    """
    metrics_dir: Optional[str] = field(
        default=None,
        metadata={
            "help": "Output directory for metrics (loss/accuracy)"
        }
    )
    results_dir: Optional[str] = field(
        default=None,
        metadata={
            "help": "Output directory for predictions"
        }
    )


@dataclass
class WindowArguments:
    """
    Arguments pertaining to output directories for metrics, results and predictions
    """
    enable_windowing: bool = field(
        default=False,
        metadata={
            'help': 'Enable windowing system alltogether'
        }
    )
    stride: Optional[int] = field(
        default=None,
        metadata={
            "help": "Stride to use when windowing features"
        }
    )
    no_answer_text: Optional[str] = field(
        default=None,
        metadata={
            "help": "Text of an unanswerable question option (Triggers "
            "label correction mechanism on windowed features)"
        }
    )


class WindowPrediction(NamedTuple):
    predictions: np.ndarray
    window_ids: List[int]
    labels: List[int]
    label: Optional[int]
    example: Optional[InputExample]


def process_ids(processor, example_ids, window_args):
    example_ids = [int(id) for id in example_ids]
    # decode id with processor
    if callable(getattr(processor, "_decode_id")):
        if window_args.enable_windowing:
            # strip truncation window index
            # ToDo: = Should join + argmax by windows
            strip = 2
        else:
            strip = 0
        # returns: context_id, question-answer_id
        decoded_ids = []
        for ex_id in example_ids:
            str_ex_id = str(ex_id)
            str_ex_id = str_ex_id[:len(str_ex_id) - strip]
            decoded_ids.append(processor._decode_id(int(str_ex_id)))
        example_ids = decoded_ids
    return example_ids


def vote_windowed_predictions(windowed_predictions):
    predictions = []
    # strategies:
    #   max along each column
    #   voting?
    for win_pred in windowed_predictions:
        predictions.append(np.argmax(win_pred.predictions, axis=0))

    return np.array(predictions)


def parse_windowed_predictions(processor, args, results, split):
    data_args = args['data_args']
    if split == Split.dev:
        examples = processor.get_dev_examples(data_args.data_dir)
    elif split == Split.test:
        examples = processor.get_test_examples(data_args.data_dir)
    else:
        raise ValueError('Saving training data is cheating!')

    predictions = []
    id_to_example_map = {example.example_id: example for example in examples}
    label_map = {label: i for i, label in enumerate(processor.get_labels())}

    example_labels = defaultdict(list)
    example_window_ids = defaultdict(list)
    example_predictions = defaultdict(list)

    zipped = zip(results.example_ids, results.label_ids, results.predictions)
    for feat_id, feat_label, feat_preds in zipped:
        str_feat_id = str(feat_id)
        example_id = int(str_feat_id[:-2])
        win_id = int(str_feat_id[-2:])
        example_labels[example_id].append(int(feat_label))
        example_window_ids[example_id].append(win_id)
        example_predictions[example_id].append(feat_preds)

    for example_id, example in id_to_example_map.items():
        windowed_predictions = WindowPrediction(
            predictions=np.vstack(example_predictions[example_id]),
            window_ids=example_window_ids[example_id],
            labels=example_labels[example_id],
            label=label_map[example.label],
            example=example,
        )
        predictions.append(windowed_predictions)

    reduced_predictions = vote_windowed_predictions(predictions)
    example_ids, label_ids = zip(*[
        (win_pred.example.example_id, win_pred.example.label)
        for win_pred in predictions
    ])
    return parse_default_predictions(
        processor=processor,
        example_ids=example_ids,
        label_ids=label_ids,
        predictions=reduced_predictions,
    )


def parse_default_predictions(processor, example_ids, label_ids, predictions):
    # ToDo := Test predictions should not have true label
    # cast to avoid json serialization issues
    example_ids = [processor._decode_id(int(ex_id)) for ex_id in example_ids]
    label_ids = [int(lab) for lab in label_ids]
    label_id_map = {i: chr(ord('A') + int(label)) for i, label in enumerate(processor.get_labels())}

    predictions = softmax(predictions, axis=1)
    predictions_dict = defaultdict(list)

    for (ex_id, q_id), true_label, preds in zip(example_ids, label_ids, predictions):
        pred_dict = {
            "probs": preds.tolist(),
            "pred_label": label_id_map[np.argmax(preds)],
            "label": label_id_map[true_label],
        }
        predictions_dict[ex_id].append(pred_dict)

    full_ids = ['-'.join([c_id, qa_id]) for c_id, qa_id in example_ids]
    predictions = np.argmax(predictions, axis=1)
    predicted_labels = [label_id_map[id] for id in predictions]
    predictions_list = dict(zip(full_ids, predicted_labels))

    return predictions_dict, predictions_list


def save_metrics(metrics, args, split):
    dir_args = args['dir_args']
    prefix = "eval" if split == Split.dev else "test"
    metrics_dict = {}
    output_metrics_file = os.path.join(
        dir_args.metrics_dir,
        f"{prefix}_metrics.json"
    )
    for key in ["eval_loss", "eval_acc"]:
        if metrics.get(key) is not None:
            metrics_dict[key] = metrics.get(key)
    if len(metrics_dict.keys()) == 0:
        logger.info("Neither loss or accuracy found on result dict!")
    else:
        with open(output_metrics_file, "w") as writer:
            writer.write(json.dumps(metrics_dict) + '\n')
        for key, value in metrics_dict.items():
            logger.info("  %s = %s", key, value)


def save_predictions(processor, results, args, split):
    dir_args, window_args = args['dir_args'], args['window_args']
    prefix = "eval" if split == Split.dev else "test"

    if window_args.enable_windowing:
        predictions_dict, predictions_list = parse_windowed_predictions(
            processor, args, results, split,
        )
    else:
        predictions_dict, predictions_list = parse_default_predictions(
            processor=processor,
            example_ids=results.example_ids,
            label_ids=results.label_ids,
            predictions=results.predictions,
        )

    output_nbest_file = os.path.join(
        dir_args.results_dir,
        f"{prefix}_nbest_predictions.json"
    )
    output_predictions_file = os.path.join(
        dir_args.results_dir,
        f"{prefix}_predictions.json"
    )

    with open(output_nbest_file, "w") as writer:
        writer.write(json.dumps(predictions_dict) + '\n')

    with open(output_predictions_file, "w") as writer:
        writer.write(json.dumps(predictions_list) + '\n')


def save_results(processor, results, args, split):
    # only predict method returns prediction outputs,
    # evaluate and train only return the metrics
    if isinstance(results, PredictionOutput):
        save_metrics(results.metrics, args, split)
        save_predictions(processor, results, args, split)
    else:
        save_metrics(results, args, split)


def main():
    # See all possible arguments in src/transformers/training_args.py
    # or by passing the --help flag to this script.
    # We now keep distinct sets of args, for a cleaner separation of concerns.
    parser = HfArgumentParser((
        ModelArguments, DataTrainingArguments,
        DirArguments, TrainingArguments, WindowArguments
    ))
    if len(sys.argv) > 1 and sys.argv[1].endswith('.json'):
        model_args, data_args, dir_args, training_args, window_args = (
            parser.parse_json_file(sys.argv[1])
        )
    else:
        model_args, data_args, dir_args, training_args, window_args = (
            parser.parse_args_into_dataclasses()
        )

    if (
        os.path.exists(training_args.output_dir)
        and [f for f in os.listdir(training_args.output_dir) if f != '.gitignore']
        and training_args.do_train
        and not training_args.overwrite_output_dir
    ):
        raise ValueError(
            f"Output directory ({training_args.output_dir}) already exists and is not empty. Use --overwrite_output_dir to overcome."
        )

    all_args = {
        'model_args': model_args,
        'data_args': data_args,
        'dir_args': dir_args,
        'training_args': training_args,
        'window_args': window_args,
    }
    # Setup logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        level=logging.INFO if training_args.local_rank in [-1, 0] else logging.WARN,
    )
    logger.warning(
        "Process rank: %s, device: %s, n_gpu: %s, distributed training: %s, 16-bits training: %s",
        training_args.local_rank,
        training_args.device,
        training_args.n_gpu,
        bool(training_args.local_rank != -1),
        training_args.fp16,
    )
    logger.info("Training/evaluation parameters %s", training_args)

    # Set seed
    set_seed(training_args.seed)

    try:
        processor = processors[data_args.task_name]()
        label_list = processor.get_labels()
        num_labels = len(label_list)
    except KeyError:
        raise ValueError("Task not found: %s" % (data_args.task_name))

    # Load pretrained model and tokenizer
    #
    # Distributed training:
    # The .from_pretrained methods guarantee that only one local process can concurrently
    # download model & vocab.

    config = AutoConfig.from_pretrained(
        model_args.config_name if model_args.config_name else model_args.model_name_or_path,
        num_labels=num_labels,
        finetuning_task=data_args.task_name,
        cache_dir=model_args.cache_dir,
    )
    tokenizer = AutoTokenizer.from_pretrained(
        model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
        cache_dir=model_args.cache_dir,
    )
    model = AutoModelForMultipleChoice.from_pretrained(
        model_args.model_name_or_path,
        from_tf=bool(".ckpt" in model_args.model_name_or_path),
        config=config,
        cache_dir=model_args.cache_dir,
    )

    # Get datasets
    train_dataset = (
        MultipleChoiceDataset(
            data_dir=data_args.data_dir,
            tokenizer=tokenizer,
            task=data_args.task_name,
            max_seq_length=data_args.max_seq_length,
            overwrite_cache=data_args.overwrite_cache,
            mode=Split.train,
            enable_windowing=window_args.enable_windowing,
            stride=window_args.stride,
            no_answer_text=window_args.no_answer_text,
        )
        if training_args.do_train
        else None
    )
    eval_dataset = (
        MultipleChoiceDataset(
            data_dir=data_args.data_dir,
            tokenizer=tokenizer,
            task=data_args.task_name,
            max_seq_length=data_args.max_seq_length,
            overwrite_cache=data_args.overwrite_cache,
            mode=Split.dev,
            enable_windowing=window_args.enable_windowing,
            stride=window_args.stride,
            no_answer_text=window_args.no_answer_text,
        )
        if training_args.do_eval
        else None
    )

    test_dataset = (
        MultipleChoiceDataset(
            data_dir=data_args.data_dir,
            tokenizer=tokenizer,
            task=data_args.task_name,
            max_seq_length=data_args.max_seq_length,
            overwrite_cache=data_args.overwrite_cache,
            mode=Split.test,
            enable_windowing=window_args.enable_windowing,
            stride=window_args.stride,
            no_answer_text=window_args.no_answer_text,
        )
        if training_args.do_predict
        else None
    )

    def compute_metrics(p: EvalPrediction) -> Dict:
        preds = np.argmax(p.predictions, axis=1)
        return {"acc": simple_accuracy(preds, p.label_ids)}

    # Initialize our Trainer
    trainer = Trainer(
        model=model,
        args=training_args,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        compute_metrics=compute_metrics,
    )

    # Training
    if training_args.do_train:
        trainer.train(
            model_path=(
                model_args.model_name_or_path
                if os.path.isdir(model_args.model_name_or_path)
                else None
            )
        )
        trainer.save_model()
        # For convenience, we also re-save the tokenizer to the same directory,
        # so that you can share your model easily on huggingface.co/models =)
        if trainer.is_world_master():
            tokenizer.save_pretrained(training_args.output_dir)

    results = {}
    if training_args.do_eval:
        logger.info("*** Evaluate ***")
        result = trainer.predict(eval_dataset)
        if trainer.is_world_master():
            save_results(
                processor, result, all_args, split=Split.dev
            )
            results['eval'] = result

    if training_args.do_predict:
        logger.info("*** Test ***")
        result = trainer.predict(test_dataset)
        if trainer.is_world_master():
            save_results(
                processor, result, all_args, split=Split.test
            )
            results['test'] = result

    return results


def _mp_fn(index):
    # For xla_spawn (TPUs)
    main()


if __name__ == "__main__":
    main()



import tqdm
import random
import logging

from typing import List, Callable, Tuple
from transformers import PreTrainedTokenizer
from mc_transformers.data_classes import InputFeatures, InputExample


logger = logging.getLogger(__name__)


try:
    import nltk
    from nltk.corpus import stopwords
    from nltk.stem import SnowballStemmer, WordNetLemmatizer
    from nltk.tokenize import word_tokenize
    # trigger download if necessary
    stopwords.words('english')
    is_nltk_available = True
except ImportError:
    is_nltk_available = False
except LookupError:
    # download extra data
    nltk.download(['stopwords', 'punkt', 'wordnet'])
    is_nltk_available = True


class TextTokenizer(object):
    stop_words = stopwords.words("english") if is_nltk_available else None
    lemmatizer = WordNetLemmatizer() if is_nltk_available else None
    stemmer = SnowballStemmer("english") if is_nltk_available else None

    def __call__(self, text):
        if is_nltk_available:
            text_tokens = word_tokenize(text.lower().strip())
            return [
                self.stemmer.stem(self.lemmatizer.lemmatize(w))
                for w in text_tokens if w not in self.stop_words
            ]
        else:
            logger.warning(
                "You are using a `TextTokenizer` for windowing but NLTK module"
                " is not installed, you can install with "
                "`pip install mc_transformers[windowing]`"
            )
            return text.lower().strip().split(' ')


def argmax(arr: List) -> int:
    return max(enumerate(arr), key=lambda elem: len(elem[1]))[0]


def match_text_by_tokenizer(
    text_a: str,
    text_b: str,
    text_tokenizer: TextTokenizer,
) -> bool:
    text_a_tokens = text_tokenizer(text_a)
    text_b_tokens = text_tokenizer(text_b)
    threshold = 0.5
    nof_correct = sum([
        1 if token in text_a_tokens else 0
        for token in text_b_tokens
    ])
    if nof_correct > 0:
        nof_correct /= len(text_b_tokens)
    return (nof_correct > threshold)


def should_correct_label(
    context: str,
    endings: str,
    no_answer_text: str,
    text_tokenizer: TextTokenizer,
) -> bool:
    # find if the answer is contained in the context,
    # which possibly indicates that the current window
    # context is enough to guess the answer
    # ToDo := This should be revised, if inference is to be made,
    # words do not necessarily appear in the options
    ctx = context.lower()
    endings = [end.lower() for end in endings]
    correct = False
    for ans in endings:
        correct = (
            (no_answer_text == ans or ctx.find(ans) != -1) or
            match_text_by_tokenizer(
                text_a=ctx, text_b=ans, text_tokenizer=text_tokenizer
            )
        )

        if correct:
            break
    return correct


def should_window(
    example: InputExample,
    tokenizer: PreTrainedTokenizer,
    max_length: int,
    no_answer_text: str,
) -> bool:
    context = example.contexts[0]
    context_tokens = tokenizer.encode(context, add_special_tokens=False)
    concats = concat_question_and_endings(
        example.question, example.endings + (
            [no_answer_text]
            if no_answer_text is not None and no_answer_text != ""
            else []
        )
    )
    longest_concat = concats[argmax(concats)]
    # get the longest span to test max length
    text_b_tokens = tokenizer.encode(longest_concat, add_special_tokens=False)
    enable_windowing = len(context_tokens) + len(text_b_tokens) > max_length
    return enable_windowing, len(text_b_tokens)


def correct_label(
    context: str,
    endings: List[str],
    no_answer_text: str,
    text_tokenizer: TextTokenizer,
) -> Tuple[int, List[str]]:
    # find `no answer text` in endings, if not found, replace a random ending
    # with the provided one and mark it as the correct answer
    found = False
    label_index = -1
    endings = [end.lower() for end in endings]
    for idx, ans in enumerate(endings):
        found = (
            (no_answer_text == ans or ans.find(no_answer_text) != -1) or
            match_text_by_tokenizer(
                text_a=no_answer_text, text_b=ans, text_tokenizer=text_tokenizer,
            )
        )
        if found:
            label_index = idx
            break

    if not found:
        label_index = random.choice(range(len(endings)))
        endings[label_index] = no_answer_text

    return label_index, endings


def create_windows(
    context: str,
    tokenizer: PreTrainedTokenizer,
    max_length: int,
    stride: int
) -> List[int]:
    context_tokens = tokenizer.encode(context, add_special_tokens=False)
    windows = []
    win_start = 0
    win_end = max_length
    total_size = len(context_tokens)
    nof_windows = round(total_size / (max_length - stride))
    for _ in range(nof_windows):
        windows.append(context_tokens[win_start:win_end])
        win_start = win_end - stride
        win_end = min(win_start + max_length, total_size)

    return [
        tokenizer.decode(
            tokens,
            skip_special_tokens=True,
            clean_up_tokenization_spaces=True
        )
        for tokens in windows
    ]


def concat_question_and_endings(question: str, endings: List[str]) -> List[str]:
    concats = []
    for end in endings:
        if question.find("_") != -1:
            # this is for cloze question
            text_b = question.replace("_", end)
        else:
            text_b = question + " " + end
        concats.append(text_b)

    return concats


def create_input_features(
    contexts: List[str],
    endings: List[str],
    example_id: int,
    label: int,
    max_length: int,
    tokenizer: PreTrainedTokenizer,
) -> InputFeatures:
    choices_inputs = []
    for text_a, text_b in zip(contexts, endings):
        inputs = tokenizer(
            text_a,
            text_b,
            add_special_tokens=True,
            max_length=max_length,
            padding="max_length",
            truncation='only_first',
            return_overflowing_tokens=True,
        )
        if "num_truncated_tokens" in inputs and inputs["num_truncated_tokens"] > 0:
            logger.info(
                "Attention! you are cropping tokens (swag task is ok). "
                "If you are training ARC and RACE and you are poping question + options,"
                "you need to try to use a bigger max seq length! "
                f"Cropped {inputs['num_truncated_tokens']} tokens"
            )

        choices_inputs.append(inputs)

    input_ids = [x["input_ids"] for x in choices_inputs]
    attention_mask = (
        [x["attention_mask"] for x in choices_inputs] if "attention_mask" in choices_inputs[0] else None
    )
    token_type_ids = (
        [x["token_type_ids"] for x in choices_inputs] if "token_type_ids" in choices_inputs[0] else None
    )
    return InputFeatures(
        example_id=example_id,
        input_ids=input_ids,
        attention_mask=attention_mask,
        token_type_ids=token_type_ids,
        label=label,
    )


def windowed_tokenization(
    example: InputExample,
    label_map: dict,
    max_window_length: int,
    max_length: int,
    stride: int,
    no_answer_text: str,
    tokenizer: PreTrainedTokenizer,
    text_tokenizer: TextTokenizer,
    window_fn: Callable = None
) -> InputFeatures:
    # ToDo := Different amount of windows will trigger an error because of
    # different size in input features? sequences should be grouped by
    # size and chopped, padded accordingly

    # ToDo := no_answer_text is not used by now, no label corrected
    able_to_correct_label = no_answer_text is not None and no_answer_text != ""
    window_fn = window_fn if window_fn is not None else create_windows
    window_texts = window_fn(
        example.contexts[0], tokenizer, max_window_length, stride
    )
    # logger.info(
    #     f"Created {len(window_texts)} windows for example "
    #     f"`{example.example_id}` with max window size {max_window_length}"
    # )
    features = []
    for win_idx, win_text in enumerate(window_texts):
        str_win_idx = str(win_idx)
        if len(str_win_idx) % 2 != 0:
            str_win_idx = '0' + str_win_idx

        if able_to_correct_label and should_correct_label(
            win_text, example.endings, no_answer_text, text_tokenizer
        ):
            label, endings = correct_label(
                win_text,
                example.endings.copy(),
                no_answer_text,
                text_tokenizer
            )
        else:
            label = label_map[example.label]
            endings = example.endings

        texts_a = [win_text] * len(endings)
        texts_b = concat_question_and_endings(example.question, endings)
        # maximum 100 windows
        example_id = int(str(example.example_id) + str_win_idx)
        features.append(
            create_input_features(
                contexts=texts_a,
                endings=texts_b,
                example_id=example_id,
                label=label,
                max_length=max_length,
                tokenizer=tokenizer,
            )
        )

    return features


def convert_examples_to_features(
    examples: List[InputExample],
    label_list: List[str],
    max_length: int,
    tokenizer: PreTrainedTokenizer,
    enable_windowing: bool = False,
    stride: int = None,
    no_answer_text: str = None,
    window_fn: Callable = None,
) -> List[InputFeatures]:
    """
    Loads a data file into a list of `InputFeatures`
    """
    if enable_windowing and stride is None:
        raise ValueError(
            'Windowing mechanism is activated, but no "stride" or '
            'was provided, please provide it or disable'
            'the mechanism altogether with `enable_windowing=False`'
        )

    features = []
    label_map = {label: i for i, label in enumerate(label_list)}
    text_tokenizer = TextTokenizer()
    # three special tokens will be added, remove them from the count
    windowing_max_length = max_length - 6
    for (ex_index, example) in tqdm.tqdm(enumerate(examples), desc="convert examples to features"):
        if ex_index % 10000 == 0:
            logger.info("Writing example %d of %d" % (ex_index, len(examples)))

        if enable_windowing:
            trigger_windowing, max_ending_length = should_window(
                example=example,
                tokenizer=tokenizer,
                max_length=windowing_max_length,
                no_answer_text=no_answer_text,
            )
        else:
            trigger_windowing, max_ending_length = False, None

        if trigger_windowing:
            features.extend(windowed_tokenization(
                example=example,
                label_map=label_map,
                max_window_length=windowing_max_length - max_ending_length,
                max_length=max_length,
                stride=stride,
                no_answer_text=no_answer_text,
                tokenizer=tokenizer,
                text_tokenizer=text_tokenizer,
                window_fn=window_fn
            ))
        else:
            # Append 00 window idx to conform to the rest of windowed features
            example_id = example.example_id
            if enable_windowing:
                example_id = int(str(example_id) + '00')

            concats = concat_question_and_endings(
                example.question, example.endings
            )
            features.append(create_input_features(
                contexts=example.contexts,
                endings=concats,
                example_id=example_id,
                label=label_map[example.label],
                max_length=max_length,
                tokenizer=tokenizer,
            ))

    for f in features[:2]:
        logger.info("*** Example ***")
        logger.info("feature: %s" % f)

    return features



from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass(frozen=True)
class InputExample:
    """
    A single training/test example for multiple choice

    Args:
        example_id: Unique id for the example.
        question: string. The untokenized text of the second sequence (question).
        contexts: list of str. The untokenized text of the first sequence (context of corresponding question).
        endings: list of str. multiple choice's options. Its length must be equal to contexts' length.
        label: (Optional) string. The label of the example. This should be
        specified for train and dev examples, but not for test examples.
    """

    example_id: Union[str, int]
    question: str
    contexts: List[str]
    endings: List[str]
    label: Optional[str]


@dataclass(frozen=True)
class InputFeatures:
    """
    A single set of features of data.
    Property names are the same names as the corresponding inputs to a model.
    """

    example_id: Union[str, int]
    input_ids: List[List[int]]
    attention_mask: Optional[List[List[int]]]
    token_type_ids: Optional[List[List[int]]]
    label: Optional[int]



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [s.strip() for s in open('requirements.txt', 'r').readlines()] 

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Guillermo E. Blanco",
    author_email='geblanco@lsi.uned.es',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
    ],
    description="Utils to run multiple choice question answering with huggingface transformers",
    entry_points={
        'console_scripts': [
            'mc_transformers=mc_transformers.mc_transformers:main',
        ],
    },
    extras_require={
        'windowing': ['nltk']
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='mc_transformers',
    name='mc_transformers',
    packages=find_packages(include=['mc_transformers', 'mc_transformers.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/geblanco/mc_transformers',
    version='0.1.8',
    zip_safe=False,
)
